from ngsolve import *
from netgen.geom2d import unit_square
from math import pi

G = sin(pi * x)
H = -G
uex = (1 - y) * G
gradu = (pi * (1 - y) * cos(pi * x), H)
F = pi * pi * uex

ngsglobals.msg_level = 0

order_flux = 2
mesh = Mesh(unit_square.GenerateMesh(maxh = 0.5))
mesh.Refine()
V = Compress(HDiv(mesh, order = order_flux, dirichlet = 'top'))
Q = Compress(L2(mesh, order = order_flux - 1))
fes = FESpace([V,Q])

sigma, u = fes.TrialFunction()
tau, v = fes.TestFunction()

normal = specialcf.normal(mesh.dim)

gf = GridFunction(fes, 'sol')
gfsigma, gfu = gf.components
gfsigma.Set(H * normal, BND)

a = BilinearForm(fes)
a += (sigma * tau + div(tau) * u + div(sigma) * v) * dx
a.Assemble()
#print(a.mat)
f = LinearForm(fes)
# The trace operator is required for H(div).
f += -F * v * dx + G * (tau.Trace() * normal) * ds ('left|bottom|right')
f.Assemble()

f.vec.data -= a.mat * gf.vec
gf.vec.data += a.mat.Inverse(freedofs = fes.FreeDofs()) * f.vec

gfuex = GridFunction(Q, 'uexact')
gfuex.Set(uex)

graduex = GridFunction(V, 'graduexact')
graduex.Set(gradu)

print("u error: ", sqrt(Integrate((gfu - uex) ** 2, mesh, order = 2)))
print("u error: ", sqrt(Integrate((gfu - gfuex) ** 2, mesh)))
print("sigma error: ", sqrt(Integrate((gfsigma - gradu) ** 2, mesh, order = 3)))
print("sigma error: ", sqrt(Integrate((gfsigma - graduex) ** 2, mesh)))
